MODULE files
  IMPLICIT NONE

CONTAINS

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! READ isat.opt file
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE read_options_file
    IMPLICIT NONE
    INTEGER :: iu,ios,iline,len_option,len_value
    CHARACTER (len=100) :: line, option, value

    CALL open_file_for_read('isat.opt',iu)
    IF (iu==0) RETURN

    iline = 0
    ios = 0
    read_lines: DO
      iline = iline + 1

      READ(iu,FMT='(A)',IOSTAT=ios) line

      ! check for end of file or max file lines
      IF (ios<0.OR.iline>1000) EXIT read_lines

      ! remove comments and replace tabs with spaces
      CALL remove_comments(line)
      CALL replace_tabs(line)

      ! check for blank line
      IF (LEN_TRIM(line)==0) CYCLE read_lines
        
      ! parse into option and value
      line = ADJUSTL(line)
      
      ! split line into option and value portions
      CALL split_str(line,' ',option,value)
      value = ADJUSTL(value)
      
      ! get lengths of strings
      len_option = LEN_TRIM(option)
      len_value = LEN_TRIM(value)

      ! check for errors
      IF ((ios>0).OR.(len_option==0).OR.(len_value==0)) THEN
         PRINT *, ''
         PRINT *, 'Invalid syntax in options file (isat.opt) at line',iline
         PRINT *, 'Line  : ',TRIM(line)
         PRINT *, 'Option: ',TRIM(option)
         PRINT *, 'Value : ',TRIM(value)
         CYCLE read_lines
      ENDIF

      ! convert to lower case
      CALL lowercase(option(1:len_option))
      CALL lowercase(value(1:len_value))

      CALL set_option(option(1:len_option),value(1:len_value))

    ENDDO read_lines

    ! close isat.opt
    CLOSE(iu)

    ! print current solver options
    CALL open_file_for_write('isat_current.opt',iu)

    WRITE(iu,*) ' ! Current options for ISAT'
    WRITE(iu,*) ' ! Template for isat.opt'

    WRITE(iu,FMT='(A35,1X,I8)') 'print_level', 0

    CLOSE(iu)

    RETURN
  END SUBROUTINE read_options_file

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! READ isat.opt file
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE set_option (option,value)
    IMPLICIT NONE

    LOGICAL :: error
    INTEGER :: ivalue,len_option,len_value,itype
    CHARACTER (len=*) :: option, value
    DOUBLE PRECISION :: dvalue

    ! initialize value type (0=string,1=real,2=integer)
    itype = 0

    ! check for real number
    dvalue = str2num (value,error)
    IF (.NOT.error) itype = 1

    ! check for integer
    ivalue = str2int (value,error)
    IF ((.NOT.error).AND.(DABS(DBLE(ivalue)-dvalue)<1.d-10)) itype = 2

    SELECT CASE (option)
    CASE('print_level')
       IF (itype==2) itype = 2  ! iparm(6)  = ivalue
    CASE DEFAULT
       PRINT *, 'Unknown option: ',TRIM(option)
    END SELECT

    RETURN
  END SUBROUTINE set_option


  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! OPEN file
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE open_file_for_read (filename,iu)
    IMPLICIT NONE
    INTEGER :: iu
    LOGICAL :: ex
    CHARACTER(*) :: filename

    ! if file open is unsuccessful, return iu = 0 and
    !   let the user decide to stop

    ! for writing, see if the file is already open
    INQUIRE(FILE=filename,EXIST=ex)
    IF (ex) THEN
       CALL get_unit(iu)
       ! open file for reading
       OPEN(UNIT=iu,FILE=filename,STATUS='UNKNOWN')
    ELSE
       PRINT *, 'File ', TRIM(filename), ' does not exist.'
       iu = 0
    ENDIF

    RETURN
  END SUBROUTINE open_file_for_read

  SUBROUTINE open_file_for_write (filename,iu)
    IMPLICIT NONE
    INTEGER :: iu
    LOGICAL :: opened_state
    CHARACTER(*) :: filename

    ! for writing, see if the file is already open
    INQUIRE(FILE=filename,OPENED=opened_state)
    IF (.NOT.opened_state) THEN
       CALL get_unit(iu)
       ! open file for reading
       OPEN(UNIT=iu,FILE=filename,STATUS='UNKNOWN')
    ELSE
       PRINT *, 'OPEN_FILE_FOR_WRITE: File ', TRIM(filename), ' is currently being used by another process.'
       CALL get_unit(iu)
       OPEN(UNIT=iu,STATUS='UNKNOWN')
       PRINT *, 'Writing to UNIT=',iu,' instead'
    ENDIF

    RETURN
  END SUBROUTINE open_file_for_write

  SUBROUTINE open_file_for_append (filename,iu)
    IMPLICIT NONE
    INTEGER :: iu
    LOGICAL :: opened_state,exists
    CHARACTER(*) :: filename

    CALL get_unit(iu)

    ! for writing, see if the file is already open
    INQUIRE(FILE=filename,OPENED=opened_state,EXIST=exists)
    IF (.NOT.opened_state.AND.exists) THEN
       ! open file for reading
       OPEN(UNIT=iu,FILE=filename,STATUS='OLD',ACCESS='APPEND')
    ELSE IF (opened_state) THEN
       PRINT *, 'OPEN_FILE_FOR_WRITE: File ', TRIM(filename), ' is currently being used by another process.'
       OPEN(UNIT=iu,STATUS='UNKNOWN')
       PRINT *, 'Writing to UNIT=',iu,' instead'
    ELSE IF (.NOT.exists) THEN
       !PRINT *, 'OPEN_FILE_FOR_APPEND: File ', TRIM(filename), ' does not exist.'
       !PRINT *, 'Creating new file.'
       OPEN(UNIT=iu,FILE=filename,STATUS='UNKNOWN')       
    ENDIF

    RETURN
  END SUBROUTINE open_file_for_append

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! CLOSE file
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE close_file (iu)
    IMPLICIT NONE
    INTEGER :: iu

    CLOSE (iu)

    RETURN
  END SUBROUTINE close_file

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! Copy file from filename1 to filename2
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE copy_file (filename1,filename2)
    IMPLICIT NONE
    INTEGER :: iu1,iu2,ios
    CHARACTER(*) :: filename1,filename2
    CHARACTER(100) :: line

    ! initialize
    ios = 0
    line = ''

    ! open filename1
    CALL open_file_for_read(filename1,iu1)
    CALL open_file_for_write(filename2,iu2)
    IF (iu1/=0) THEN
       DO WHILE (ios>=0)
          READ(iu1,FMT='(A)',IOSTAT=ios) line
          IF (ios>=0) THEN
             WRITE(iu2,FMT='(A)') TRIM(line)
          ENDIF
       ENDDO
       CLOSE(iu1)
    ELSE
       PRINT *, 'File copy failed: ', TRIM(filename1), ' does not exist'
       ! write one blank line for filename2
       WRITE (iu2,*) ' '
    ENDIF

    CLOSE(iu2)

    RETURN
  END SUBROUTINE copy_file

  ! ***************************************
  ! get file unit
  ! ***************************************
  SUBROUTINE get_unit(iu)
    IMPLICIT NONE
    INTEGER :: i,iu
    LOGICAL :: io
    do_i: DO i = 10,1000
       INQUIRE(UNIT=i,OPENED=io)
       IF (.NOT.io) THEN
          iu = i
          EXIT do_i
       END IF
    END DO do_i
    RETURN
  END SUBROUTINE get_unit

  SUBROUTINE replace_tabs (str)
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    ! Transform tab characters in str into spaces
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    IMPLICIT NONE
    CHARACTER (len=*) :: str
    INTEGER           :: j,k
    CHARACTER (len=1) :: replace
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    replace = ACHAR(9) ! // ACHAR(10)... can add others to the list to replace
    DO j=1,LEN_TRIM(str)
       k = INDEX(replace,str(j:j))
       IF (k > 0) str(j:j) = ' '
    END DO
  END SUBROUTINE replace_tabs

  ! ***************************************
  ! replace instances of str1 by str2 
  ! ***************************************
  SUBROUTINE replace_str (str,str1,str2)
    IMPLICIT NONE
    CHARACTER(*) :: str,str1,str2
    LOGICAL :: remove_spaces
    INTEGER :: i,count,match,start
    INTEGER :: max_str,len_str1,len_str2,d12

    match = 0
    start = 1
    max_str = LEN(str)
    len_str1 = LEN(str1)
    len_str2 = LEN(str2)
    IF (str1==' '.AND.str2=='') THEN
       remove_spaces = .TRUE.
    ELSE
       remove_spaces = .FALSE.
    ENDIF
    ! every replacement will increase str by this many characters
    d12 = MAX(0,len_str2-len_str1)

    ! replace instances of str1 with temp
    count = 0
    replace_it: DO
       count = count + 1
       i = INDEX(str(start:),str1)
       ! compensate for offset
       IF (i>0) i = i + start - 1

       ! exit if str1 not found (i==0) or number of matches>=1000 or string length violation
       IF (i<=0.OR.count>=1000) EXIT replace_it

       ! check for completion of space replacement
       IF (remove_spaces) THEN
         IF (i>LEN_TRIM(str)) EXIT replace_it
       ENDIF

       ! found at beginning of string
       IF (i==1) THEN
          IF (len_str2<=max_str) THEN
             str = str2//str(len_str1+1:max_str-d12)
             match = match + 1
             start = len_str2+1
          ELSE
             str = str2(1:max_str)
             match = match + 1
             EXIT replace_it
          ENDIF
       END IF

       ! found in the middle of the string
       IF (i>=2) THEN
          IF (i-1+len_str2<=max_str) THEN
             str = str(1:i-1)//str2//str(i+len_str1:max_str-d12)
             match = match + 1
             start = i+len_str2
          ELSE
             str = str(1:i-1)//str2(1:max_str-i+1)
             match = match + 1
             EXIT replace_it
          ENDIF
       END IF
    END DO replace_it

    RETURN
  END SUBROUTINE replace_str

  ! ***************************************
  ! remove blanks
  ! ***************************************
  SUBROUTINE remove_blanks (str)
    IMPLICIT NONE
    CHARACTER(*) :: str
    INTEGER :: i,j,len_str

    str = ADJUSTL(str)
    len_str = LEN_TRIM(str)

    i = 1
    j = 1
    DO WHILE (j<len_str)
       IF (str(i:i)==' ') THEN
          str = str(1:i-1)//str(i+1:len_str)
       ELSE
          i = i+1
       ENDIF
       j = j + 1
    ENDDO

    RETURN
  END SUBROUTINE remove_blanks

  ! ***************************************
  ! count number of characters in string
  ! ***************************************
  SUBROUTINE count_char (str,char,n_char)
    IMPLICIT NONE
    CHARACTER :: char
    CHARACTER(*) :: str
    INTEGER :: i,j,len_str,n_char

    str = ADJUSTL(str)
    len_str = LEN_TRIM(str)

    j = 0
    DO i = 1,len_str
       IF (str(i:i)==char) j = j + 1
    ENDDO

    n_char = j

    RETURN
  END SUBROUTINE count_char

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! Convert string to real number
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE num2str (num,str)
    IMPLICIT NONE
    CHARACTER (len=*) :: str   
    DOUBLE PRECISION  :: num

    str = 'NaN'
    IF (DABS(num)>1.d5) WRITE(str,FMT='(ES10.3)') num
    IF (DABS(num)<=1.d5.AND.DABS(num)>1.d3) WRITE(str,FMT='(F10.1)') num
    IF (DABS(num)<=1.d3.AND.DABS(num)>0.5d0) WRITE(str,FMT='(F10.3)') num
    IF (DABS(num)<=0.5d0.AND.DABS(num)>=1.d-3) WRITE(str,FMT='(F10.5)') num
    IF (DABS(num)<=1.d-3.AND.DABS(num)>1.d-99) WRITE(str,FMT='(ES10.3)') num
    IF (DABS(num)<=1.d-99) str = '0.00'

    ! get rid of + sign (not good for php)
    CALL replace_str (str,'+','')

    RETURN
  END SUBROUTINE num2str

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! Convert string to real number
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  FUNCTION str2num (str,error) RESULT (res)
    IMPLICIT NONE
    LOGICAL           :: error
    INTEGER           :: istat
    CHARACTER (len=*) :: str   
    DOUBLE PRECISION  :: res

    READ(str,*,IOSTAT=istat) res
    error = istat /= 0

    RETURN
  END FUNCTION str2num

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! Convert string to integer
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  FUNCTION str2int (str,error) RESULT (ires)
    IMPLICIT NONE
    LOGICAL           :: error
    INTEGER           :: istat,ires
    CHARACTER (len=*) :: str   

    READ(str,*,IOSTAT=istat) ires
    error = istat /= 0

    IF (error) ires = 0

    RETURN
  END FUNCTION str2int


  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! Convert integer to string
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE int2str (i,str)
    IMPLICIT NONE
    LOGICAL           :: error
    INTEGER           :: istat,i
    CHARACTER (len=*) :: str   
    DOUBLE PRECISION  :: res
    WRITE(str,FMT='(I10)',IOSTAT=istat) i
    error = istat /= 0
    IF (error) THEN
       str = ''
    ELSE
       str = ADJUSTL(str)
    ENDIF
    RETURN
  END SUBROUTINE int2str


  SUBROUTINE lowercase (str)
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    ! Transform upper case letters in str into lower case
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    IMPLICIT NONE
    CHARACTER (LEN=*)              :: str
    INTEGER                        :: j,k
    CHARACTER (LEN=*),   PARAMETER :: lc = 'abcdefghijklmnopqrstuvwxyz'
    CHARACTER (LEN=*),   PARAMETER :: uc = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    DO j=1,LEN_TRIM(str)
       k = INDEX(uc,str(j:j))
       IF (k > 0) str(j:j) = lc(k:k)
    END DO
  END SUBROUTINE lowercase

  SUBROUTINE uppercase (str)
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    ! Transform upper case letters in str into lower case
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    IMPLICIT NONE
    CHARACTER (LEN=*)              :: str
    INTEGER                        :: j,k
    CHARACTER (LEN=*),   PARAMETER :: lc = 'abcdefghijklmnopqrstuvwxyz'
    CHARACTER (LEN=*),   PARAMETER :: uc = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    !----- -------- --------- --------- --------- --------- --------- --------- -------
    DO j=1,LEN_TRIM(str)
       k = INDEX(lc,str(j:j))
       IF (k > 0) str(j:j) = uc(k:k)
    END DO
  END SUBROUTINE uppercase

  ! split string str into str1 and str2, delimited by del
  SUBROUTINE split_str(str,del,str1,str2)
    IMPLICIT NONE
    CHARACTER(*) :: str,del,str1,str2
    INTEGER :: i

    i = INDEX(str,del)

    IF (i<=1) THEN
       str1 = str
       str2 = ''
    END IF
    IF (i>=1) THEN
       str1 = str(1:i-1)
       str2 = str(i+1:)
    END IF

    RETURN
  END SUBROUTINE split_str

  ! remove comments
  SUBROUTINE remove_comments(str)
    IMPLICIT NONE
     ! comment characters
    CHARACTER (LEN=1), DIMENSION(4), PARAMETER :: comment = (/ '!','#','%','' /)
    CHARACTER (*) :: str
    INTEGER :: i,j,clip
    ! comment characters !,#,%,
    clip = 0
    DO i = 1,4
       j = INDEX(str,comment(i))
       IF (j>0) THEN
          IF (clip>0) THEN
             ! take minimum position
             clip = MIN(clip,j)
          ELSE
             ! initialize position
             clip = j
          ENDIF
       ENDIF
    ENDDO
    IF (clip>0) str(clip:) = ' '
    RETURN
  END SUBROUTINE remove_comments

  ! read line of file
  SUBROUTINE read_line(iu,line,ios)
    IMPLICIT NONE
    CHARACTER(*) :: line
    INTEGER :: iu,ios,len_line

    ! ios==0 with a good read
    READ(iu,FMT='(A)',IOSTAT=ios) line
    line = ADJUSTL(line)
    len_line = LEN_TRIM(line)

    CALL lowercase(line(1:len_line))
    CALL replace_tabs(line(1:len_line))

    RETURN
  END SUBROUTINE read_line

  ! count number of lines in a file
  FUNCTION file_lines (filename) RESULT (lines)
    IMPLICIT NONE
    CHARACTER(*) :: filename
    CHARACTER(10) :: line
    INTEGER :: lines, ios, iu
    LOGICAL :: ex

    iu = 0
    lines = 0

    ! for reading, see if the filename exists
    INQUIRE(FILE=filename,EXIST=ex)
    IF(ex) THEN
       CALL get_unit(iu)
       ! open file for reading
       OPEN(UNIT=iu,FILE=filename,STATUS='UNKNOWN')

       ! read until no records exist
       ios = 0
       ! ios = 0 good read
       ! ios > 0 problem with read
       ! ios < 0 end of file reached
       DO WHILE (ios>=0)
          READ(iu,*,IOSTAT=ios) line
          lines = lines + 1
       END DO

       ! last line read was end of file
       lines = lines - 1

       ! close file
       CLOSE(iu)
    ELSE
       !PRINT *, 'file_lines: ',TRIM(filename),' does not exist'
    END IF

    RETURN
  END FUNCTION file_lines

  ! count number of columns in the first row of a delimited file (e.g. csv)
  FUNCTION file_columns (filename,delim) RESULT (cols)
    IMPLICIT NONE
    CHARACTER(*) :: filename,delim
    INTEGER, PARAMETER :: max_width = 8000
    CHARACTER(LEN=max_width) :: line
    INTEGER :: i,m,n
    INTEGER :: cols,ios,iu
    LOGICAL :: skip

    iu = 0

    CALL open_file_for_read(filename,iu)
    READ(iu,FMT='(A)',IOSTAT=ios) line

    cols = 1
    n = LEN_TRIM(line)
    m = 1 !LEN(delim)

    skip = .FALSE.

    IF (m==0) THEN
       PRINT *, 'file_columns_error: zero length for delimiter'
       PRINT *, 'STOPPING. . .'
       STOP
    ENDIF

    DO i = 1,n
       IF (line(i:i+m-1)==delim) THEN
          IF (.NOT.skip) cols = cols + 1
          ! skip successive delimiters (e.g. multiple spaces)
          skip = .TRUE.
       ELSE
          skip = .FALSE.
       ENDIF
    ENDDO

    CLOSE (iu)

    RETURN
  END FUNCTION file_columns

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! ********************************************************************
  ! Read CSV file into storage arrays
  ! ********************************************************************
  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  SUBROUTINE read_csv (filename,m,n,csv_data)
    IMPLICIT NONE
    CHARACTER :: delim
    CHARACTER(*) :: filename
    INTEGER, PARAMETER :: max_width = 8000
    CHARACTER (len=max_width) :: line,str1,str2
    CHARACTER (len=50), DIMENSION(:), ALLOCATABLE :: sdata,csv_header
    DOUBLE PRECISION, DIMENSION(:,:) :: csv_data
    LOGICAL, DIMENSION(:,:), ALLOCATABLE :: csv_data_error
    INTEGER :: i,j,m,n
    INTEGER :: ios,iu
    LOGICAL :: chk_open,error,header

    header = .false.
    error = .false.
    iu = 0
    str1 = ''
    str2 = ''

    ! check CSV file for opened state
    INQUIRE(FILE=filename,OPENED=chk_open)
    IF (chk_open) THEN
       PRINT *, 'read_csv: file ', TRIM(filename), ' is in use by'
       PRINT *, '  another process.'
       PRINT *, 'STOPPING. . .'
       STOP
    ENDIF

    delim = ','
    m = file_lines(filename) - 1
    n = file_columns(filename,delim)

    IF (n<=0.OR.m<=0) THEN
       PRINT *, ''
       PRINT *, 'no data in ', TRIM(filename)
       PRINT *, 'STOPPING. . .'
       STOP
    ENDIF

    ! initialize CSV info storage
    IF (header) THEN
      ALLOCATE (csv_header(n))
      DO i = 1,n
         csv_header(i) = '*****'
      ENDDO
    ENDIF
    ALLOCATE (csv_data_error(m,n))
    ALLOCATE (sdata(n))

    ! default data for mis-reads
    DO i = 1,n
       DO j = 1,m
          csv_data(j,i) = 0.d0
       ENDDO
    ENDDO

    CALL open_file_for_read(filename,iu)

    IF (header) THEN
      ! read header
      READ(iu,FMT='(A)',IOSTAT=ios) line

      ! parse header
      IF (n==1) THEN
         csv_header(1) = TRIM(line)
      ELSE
         DO i = 1,n-1
            CALL split_str (line,delim,str1,str2)
            CALL lowercase (str1)
            csv_header(i) = TRIM(str1)
            line = str2
         ENDDO
         csv_header(n) = TRIM(str2)
      ENDIF
    ENDIF

    ! read data as string
    DO i = 1,m
       READ(iu,*,IOSTAT=ios) (sdata(j), j=1,n)
       IF (ios==0) THEN
          DO j = 1,n
             csv_data(i,j) = str2num (sdata(j),error)
             csv_data_error(i,j) = error
          ENDDO
       ELSE
          PRINT *, 'read_csv: error in reading row', i
          PRINT *, 'STOPPING...'
          STOP
       ENDIF
    END DO
    CLOSE (iu)

    DEALLOCATE (sdata)

    RETURN
  END SUBROUTINE read_csv

  SUBROUTINE write_csv (filename,iu,m,n,csv_data)
    IMPLICIT NONE

    LOGICAL :: header
    INTEGER :: iu,i,j,m,n
    CHARACTER :: delim = ','
    CHARACTER (len=*) :: filename
    ! CHARACTER (len=*), DIMENSION(*), OPTIONAL :: csv_header
    DOUBLE PRECISION, DIMENSION(:,:) :: csv_data

    header = .FALSE.

    ! write data to CSV file
    CALL open_file_for_write(filename,iu)

!!$    IF (header) THEN
!!$      ! write header
!!$      DO i = 1,n-1
!!$         WRITE(iu,FMT='(1X,A,A)',ADVANCE='NO') TRIM(csv_header(i)),TRIM(delim)
!!$      ENDDO
!!$      ! last entry has no delimiter
!!$      WRITE(iu,FMT='(1X,A)',ADVANCE='YES') TRIM(csv_header(n))
!!$    ENDIF

    ! write data
    DO i = 1,m
       DO j = 1,n-1
          WRITE(iu,FMT='(1X,ES12.4,A)',ADVANCE='NO') csv_data(i,j),TRIM(delim)
       ENDDO
       ! last entry has no delimiter
       WRITE(iu,FMT='(1X,ES12.4)',ADVANCE='YES') csv_data(i,n)
    END DO

    CLOSE (iu)

    RETURN
  END SUBROUTINE write_csv

  ! ***************************************
  ! determine if a file exists
  ! ***************************************
  FUNCTION exists (filename) RESULT (ex)
    IMPLICIT NONE
    CHARACTER(*) :: filename
    LOGICAL :: ex

    INQUIRE(FILE=filename,EXIST=ex)

    RETURN
  END FUNCTION exists

END MODULE files
